#' Pipeline for Single Cell Data
#'
#' @md
#'
#' @description
#' Semi-supervised isoform detection and annotation for long read data.
#' This variant is for single cell data. By default, this pipeline demultiplexes input
#' fastq data (\code{match_cell_barcode = TRUE}). Specific parameters relating to
#' analysis can be changed either through function arguments, or through a
#' configuration JSON file.
#'
#' @details
#' By default FLAMES use minimap2 for read alignment. After the genome alignment step (\code{do_genome_align}), FLAMES summarizes the alignment for each read by grouping reads
#' with similar splice junctions to get a raw isoform annotation (\code{do_isoform_id}). The raw isoform
#' annotation is compared against the reference annotation to correct potential splice site
#' and transcript start/end errors. Transcripts that have similar splice junctions
#' and transcript start/end to the reference transcript are merged with the
#' reference. This process will also collapse isoforms that are likely to be truncated
#' transcripts. If \code{isoform_id_bambu} is set to \code{TRUE}, \code{bambu::bambu} will be used to generate the updated annotations.
#' Next is the read realignment step (\code{do_read_realign}), where the sequence of each transcript from the update annotation is extracted, and
#' the reads are realigned to this updated \code{transcript_assembly.fa} by minimap2. The
#' transcripts with only a few full-length aligned reads are discarded.
#' The reads are assigned to transcripts based on both alignment score, fractions of
#' reads aligned and transcript coverage. Reads that cannot be uniquely assigned to
#' transcripts or have low transcript coverage are discarded. The UMI transcript
#' count matrix is generated by collapsing the reads with the same UMI in a similar
#' way to what is done for short-read scRNA-seq data, but allowing for an edit distance
#' of up to 2 by default. Most of the parameters, such as the minimal distance to splice site and minimal percentage of transcript coverage
#' can be modified by the JSON configuration file (\code{config_file}).
#'
#' @param annot The file path to gene annotations file in gff3  format
#' @param fastq The file path to input fastq file
#' @param in_bam Optional file path to a bam file to use instead of fastq file (skips initial alignment step)
#' @param outdir The path to directory to store all output files.
#' @param genome_fa The file path to genome fasta file.
#' @param minimap2_dir Path to the directory containing minimap2, if it is not in PATH. Only required if either or both of
#' \code{do_genome_align} and \code{do_read_realign} are \code{TRUE}.
#' @param seed Integer or NULL; The random seed for minimap2.
#' @param downsample_ratio Integer; downsampling ratio if performing downsampling analysis.
#' @param config_file File path to the JSON configuration file. If specified, \code{config_file} overrides
#' all configuration parameters
#' @param do_genome_align Boolean; specifies whether to run the genome alignment step. \code{TRUE} is recommended
#' @param do_isoform_id Boolean; specifies whether to run the isoform identification step. \code{TRUE} is recommended
#' @param isoform_id_bambu Boolean; specifies whether to use Bambu for isoform identification.
#' @param do_read_realign Boolean; specifies whether to run the read realignment step. \code{TRUE} is recommended
#' @param do_transcript_quanti Boolean; specifies whether to run the transcript quantification step. \code{TRUE} is recommended
#' @param gen_raw_isoform Boolean; specifies whether a gff3 should be generated containing the raw isoform information in the isoform identification step
#' @param has_UMI Boolean; specifies if the data contains UMI.
#' @param MAX_DIST Real; maximum distance allowed when merging splicing sites in isoform consensus clustering.
#' @param MAX_TS_DIST Real; maximum distance allowed when merging transcript start/end position in isoform consensus clustering.
#' @param MAX_SPLICE_MATCH_DIST Real; maximum distance allowed when merging splice site called from the data and the reference annotation.
#' @param min_fl_exon_len Real; minimum length for the first exon outside the gene body in reference annotation. This is to correct the alignment artifact
#' @param Max_site_per_splice Real; maximum transcript start/end site combinations allowed per splice chain
#' @param Min_sup_cnt Real; minimum number of read support an isoform. Decreasing this number will significantly increase the number of isoform detected.
#' @param Min_cnt_pct Real; minimum percentage of count for an isoform relative to total count for the same gene.
#' @param Min_sup_pct Real; minimum percentage of count for an splice chain that support a given transcript start/end site combination.
#' @param strand_specific 1, -1 or 0. 1 indicates if reads are in the same
#' strand as mRNA, -1 indicates reads are reverse complemented, 0 indicates
#' reads are not strand specific.
#' @param remove_incomp_reads Real; determines the strength of truncated isoform filtering. Larger number means more stringent filtering.
#' @param use_junctions Boolean; determiens whether to use known splice junctions to help correct the alignment results
#' @param no_flank Boolean; passed to minimap2 for synthetic spike-in data. Refer to Minimap2 document for more details
#' @param use_annotation Boolean; specifies whether to use reference to help annotate known isoforms
#' @param min_tr_coverage Real; minimum percentage of isoform coverage for a read to be aligned to that isoform
#' @param min_read_coverage Real; minimum percentage of read coverage for a read to be uniquely aligned to that isoform
#' @param UMI_LEN Integer; the length of UMI sequence in bases
#' @param reference_csv The file path to the reference csv used for demultiplexing
#' @param match_barcode Boolean; specifies if demultiplexing should be performed using `FLAMES::match_cell_barcode_cpp`
#' @return \code{sc_long_pipeline} returns a SingleCellExperiment object, containing a count
#' matrix as an assay, gene annotations under metadata, as well as a list of the other
#' output files generated by the pipeline. The pipeline also outputs a number of output
#' files into the given \code{outdir} directory. These output files generated by the pipeline are:
#' \itemize{
#'  \item{transcript_count.csv.gz}{ - a transcript count matrix (also contained in the SingleCellExperiment)}
#'  \item{isoform_annotated.filtered.gff3}{ - isoforms in gff3 format (also contained in the SingleCellExperiment)}
#'  \item{transcript_assembly.fa}{ - transcript sequence from the isoforms}
#'  \item{align2genome.bam}{ - sorted BAM file with reads aligned to genome}
#'  \item{realign2transcript.bam}{ - sorted realigned BAM file using the transcript_assembly.fa as reference}
#'  \item{tss_tes.bedgraph}{ - TSS TES enrichment for all reads (for QC)}
#' }
#'
#' @details The default parameters can be changed either through the function
#' arguments are through the configuration JSON file \code{config_file}. the \code{pipeline_parameters}
#' section specifies which steps are to be executed in the pipeline - by default, all
#' steps are executed. The \code{isoform_parameters} section affects isoform detection - key
#' parameters include:
#' \itemize{
#'  \item{\code{Min_sup_cnt}}{ which causes transcripts with less reads aligned than
#' it's value to be discarded}
#'  \item{\code{MAX_TS_DIST}}{ which merges transcripts with the same intron
#' chain and TSS/TES distace less than \code{MAX_TS_DIST}}
#'  \item{\code{strand_specific}}{ which specifies if reads are in the same strand as the mRNA (1),
#' or the reverse complemented (-1) or not strand specific (0), which results in
#' strand information being based on reference annotation.}
#' }
#'
#' @seealso
#' [bulk_long_pipeline()] for bulk long data,
#' [SingleCellExperiment()] for how data is outputted
#'
#' @importFrom dplyr group_by summarise_at slice_max filter
#' @importFrom magrittr "%>%"
#' @importFrom SingleCellExperiment SingleCellExperiment reducedDimNames logcounts
#' @importFrom SummarizedExperiment rowData colData rowData<- colData<- rowRanges rowRanges<-
#' @importFrom BiocGenerics cbind colnames rownames start end
#' @importFrom utils read.csv read.table
#'
#' @example inst/examples/pipeline_example.R
#' @export
sc_long_pipeline <-
    function(annot,
             fastq,
             in_bam = NULL,
             outdir,
             genome_fa,
             minimap2_dir = "",
             seed = NULL,
             downsample_ratio = 1,
             reference_csv,
             match_barcode = TRUE,
             config_file = NULL,
             do_genome_align = TRUE,
             do_isoform_id = TRUE,
             isoform_id_bambu = FALSE,
             do_read_realign = TRUE,
             do_transcript_quanti = TRUE,
             gen_raw_isoform = TRUE,
             has_UMI = FALSE,
             UMI_LEN = 10,
             MAX_DIST = 10,
             MAX_TS_DIST = 100,
             MAX_SPLICE_MATCH_DIST = 10,
             min_fl_exon_len = 40,
             Max_site_per_splice = 3,
             Min_sup_cnt = 10,
             Min_cnt_pct = 0.01,
             Min_sup_pct = 0.2,
             strand_specific = 1,
             remove_incomp_reads = 5,
             use_junctions = TRUE,
             no_flank = TRUE,
             use_annotation = TRUE,
             min_tr_coverage = 0.75,
             min_read_coverage = 0.75) {
        checked_args <- check_arguments(
            annot,
            fastq,
            in_bam,
            outdir,
            genome_fa,
            minimap2_dir,
            downsample_ratio,
            config_file,
            do_genome_align,
            do_isoform_id = TRUE,
            isoform_id_bambu,
            do_read_realign,
            do_transcript_quanti,
            gen_raw_isoform,
            has_UMI,
            MAX_DIST,
            MAX_TS_DIST,
            MAX_SPLICE_MATCH_DIST,
            min_fl_exon_len,
            Max_site_per_splice,
            Min_sup_cnt,
            Min_cnt_pct,
            Min_sup_pct,
            strand_specific,
            remove_incomp_reads,
            use_junctions,
            no_flank,
            use_annotation,
            min_tr_coverage,
            min_read_coverage
        )

        config_file <- checked_args$config

        infq <- NULL
        if (match_barcode) {
            if (!file.exists(reference_csv)) {
                stop("reference_csv must exists.")
            }
            infq <-
                paste(outdir, "matched_reads.fastq.gz", sep = "/")
            bc_stat <-
                paste(outdir, "matched_barcode_stat", sep = "/")
            match_cell_barcode_cpp(
                fastq,
                bc_stat,
                infq,
                reference_csv,
                MAX_DIST,
                UMI_LEN
            )
        } else {
            infq <- fastq
        } # requesting to not match barcodes implies `fastq` has already been run through the
        # function in a previous FLAMES call

        out_files <-
            generic_long_pipeline(
                annot,
                infq,
                in_bam,
                outdir,
                genome_fa,
                minimap2_dir,
                seed,
                downsample_ratio,
                config_file,
                do_genome_align,
                do_isoform_id,
                isoform_id_bambu,
                do_read_realign,
                do_transcript_quanti,
                gen_raw_isoform,
                has_UMI,
                MAX_DIST,
                MAX_TS_DIST,
                MAX_SPLICE_MATCH_DIST,
                min_fl_exon_len,
                Max_site_per_splice,
                Min_sup_cnt,
                Min_cnt_pct,
                Min_sup_pct,
                strand_specific,
                remove_incomp_reads,
                use_junctions,
                no_flank,
                use_annotation,
                min_tr_coverage,
                min_read_coverage
            )

        load_genome_anno <- rtracklayer::import(annot, feature.type = c("exon", "utr"))
        sce <- generate_sc_singlecell(out_files, load_genome_anno = load_genome_anno)

        sce
    }

generate_sc_sce <- function(out_files, load_genome_anno = NULL, create_function) {
    # this method requires testing using single cell data
    mdata <- list(
        "OutputFiles" = out_files
    )

    transcript_count <- read.csv(out_files$counts, stringsAsFactors = FALSE)
    if ("fsm_annotation" %in% names(out_files)) {
        isoform_FSM_annotation <- read.csv(out_files$fsm_annotation, stringsAsFactors = FALSE)
    } else {
        isoform_FSM_annotation <- read.csv(file.path(out_files$outdir, "isoform_FSM_annotation.csv"), stringsAsFactors = FALSE)
    }

    isoform_FSM_annotation <- isoform_FSM_annotation[match(transcript_count$transcript_id, isoform_FSM_annotation$transcript_id), ]
    # transcript_count <- transcript_count[match(isoform_FSM_annotation$transcript_id, transcript_count$transcript_id), ]
    transcript_count$FSM_match <- isoform_FSM_annotation$FSM_match
    if (!all(transcript_count$transcript_id %in% isoform_FSM_annotation$transcript_id)) {
        message("Some transcript_ids are not recorded in isoform_FSM_annotation.csv")
        transcript_count$FSM_match[is.na(transcript_count$FSM_match)] <- transcript_count$transcript_id[is.na(transcript_count$FSM_match)]
    }

    cell_bcs <- colnames(transcript_count)[!(colnames(transcript_count) %in% c("transcript_id", "gene_id", "FSM_match"))]
    tr_anno <- transcript_count[, c("transcript_id", "gene_id", "FSM_match")]

    # sum transcript (FSM) counts
    mer_tmp <- transcript_count %>%
        group_by(FSM_match) %>%
        summarise_at(cell_bcs, sum)

    # Create long read SCE
    tr_anno <- tr_anno[match(mer_tmp$FSM_match, tr_anno$FSM_match), ]
    tr_sce <- create_function(
        assays = list(counts = as.matrix(mer_tmp[, -1])),
        metadata = mdata
    )
    rownames(tr_sce) <- mer_tmp$FSM_match

    isoform_gff <- rtracklayer::import.gff3(out_files$isoform_annotated)
    isoform_gff$Parent <- as.character(isoform_gff$Parent)
    isoform_gff$transcript_id <- unlist(lapply(strsplit(isoform_gff$Parent, split = ":"), function(x) {
        x[2]
    }))
    isoform_gff <- S4Vectors::split(isoform_gff, isoform_gff$transcript_id)
    missing_tr <- !(tr_anno$transcript_id %in% names(isoform_gff))

    if (!is.null(load_genome_anno)) {
        genome_anno <- S4Vectors::split(load_genome_anno, load_genome_anno$transcript_id)
        add_from_genome_anno <- (missing_tr & (tr_anno$transcript_id %in% names(genome_anno)))
        rowRanges(tr_sce[add_from_genome_anno, ]) <- genome_anno[tr_anno[add_from_genome_anno, "transcript_id"]]
        if (!all((!missing_tr) | tr_anno[, "transcript_id"] %in% names(genome_anno))) {
            message("Warning: some transcript_id could not be found in annotation file\n")
        }
    } else if (any(missing_tr)) {
        message("Warning: some transcript_id could not be found in annotation file\n")
    }

    rowRanges(tr_sce[!missing_tr, ]) <- isoform_gff[tr_anno[!missing_tr, "transcript_id"]]


    rowData(tr_sce) <- DataFrame(tr_anno)
    # return the created singlecellexperiment
    return(tr_sce)
}

generate_sc_singlecell <- function(out_files, load_genome_anno = NULL) {
    return(generate_sc_sce(out_files = out_files, load_genome_anno = load_genome_anno, create_function = SingleCellExperiment::SingleCellExperiment))
}

generate_bulk_summarized <- function(out_files, load_genome_anno = NULL) {
    return(generate_sc_sce(out_files = out_files, load_genome_anno = load_genome_anno, create_function = SummarizedExperiment::SummarizedExperiment))
}


#' Create \code{SingleCellExperiment} object from \code{FLAMES} output folder
#' @param outdir The folder containing \code{FLAMES} output files
#' @param annot (Optional) the annotation file that was used to produce the output files
#' @return a \code{SingleCellExperiment} object
#' @example inst/examples/pipeline_example.R
#' @export
create_sce_from_dir <- function(outdir, annot = NULL) {
    out_files <- list(
        counts = file.path(outdir, "transcript_count.csv.gz"),
        isoform_annotated = file.path(outdir, "isoform_annotated.gff3"),
        outdir = outdir,
        transcript_assembly = file.path(outdir, "transcript_assembly.fa"),
        align_bam = file.path(outdir, "align2genome.bam"),
        realign2transcript = file.path(outdir, "realign2transcript.bam"),
        tss_tes = file.path(outdir, "tss_tes.bedgraph")
    )
    if (!is.null(annot)) {
        out_files[["annot"]] <- annot
        load_genome_anno <- rtracklayer::import(annot, feature.type = c("exon", "utr"))
        return(generate_sc_singlecell(out_files, load_genome_anno = load_genome_anno))
    } else {
        return(generate_sc_singlecell(out_files))
    }
}

#' Create \code{SummarizedExperiment} object from \code{FLAMES} output folder
#' @param outdir The folder containing \code{FLAMES} output files
#' @param annot (Optional) the annotation file that was used to produce the output files
#' @return a \code{SummarizedExperiment} object
#' @example inst/examples/pipeline_example.R
#' @export
create_se_from_dir <- function(outdir, annot) {
    out_files <- list(
        counts = file.path(outdir, "transcript_count.csv.gz"),
        isoform_annotated = file.path(outdir, "isoform_annotated.gff3"),
        outdir = outdir,
        transcript_assembly = file.path(outdir, "transcript_assembly.fa"),
        align_bam = file.path(outdir, "align2genome.bam"),
        realign2transcript = file.path(outdir, "realign2transcript.bam"),
        tss_tes = file.path(outdir, "tss_tes.bedgraph")
    )
    if (!is.null(annot)) {
        out_files[["annot"]] <- annot
        load_genome_anno <- rtracklayer::import(annot, feature.type = c("exon", "utr"))
        return(generate_bulk_summarized(out_files, load_genome_anno = load_genome_anno))
    } else {
        return(generate_bulk_summarized(out_files))
    }
}